# https://blog.shikoan.com/imagenet-preprocessing/

import cv2
import numpy as np

class Mirror:
    def __init__( self ):
        pass

    @property
    def nb_aug( self ):
        return 2

    def __call__( self, imgs ):
        if( type(imgs) is not list ):
            imgs = [imgs]
        aug = []
        for img in imgs:
            aug.append( img )
            aug.append( np.fliplr( img ) )
        return aug

class TenCrop:
    def __init__( self, scales=[224, 256, 384, 480, 640], mirror=True, img_size=(224,224) ): # img_size=(width,height)
        self.scales = scales
        self.mirror = mirror
        self.__img_size = img_size

    @property
    def img_size( self ):
        return self.__img_size

    @property
    def nb_aug( self ):
        r = len(self.scales)*5
        if( self.mirror ):
            r *= 2
        return r

    def __call__( self, imgs ):
        if( type(imgs) is not list ):
            imgs = [imgs]
        aug = []
        for img in imgs:
            for scale in self.scales:
                if( img.shape[0] < img.shape[1] ): # height < width
                    h = scale
                    w = int(img.shape[1] * scale / img.shape[0])
                else:
                    h = int(img.shape[0] * scale / img.shape[1])
                    w = scale
                _img = cv2.resize( img, (w,h) )
                # center
                h0 = int(_img.shape[0] / 2 - self.img_size[1]/2)
                h1 = h0 + self.img_size[1]
                w0 = int(_img.shape[1] / 2 - self.img_size[0]/2)
                w1 = w0 + self.img_size[0]
                aug.append( _img[h0:h1, w0:w1, :] )
                # top left
                aug.append( _img[:self.img_size[1], :self.img_size[0], :] )
                # top right
                aug.append( _img[:self.img_size[1], -self.img_size[0]:, :] )
                # bottom left
                aug.append( _img[-self.img_size[1]:, :self.img_size[0], :] )
                # bottom right
                aug.append( _img[-self.img_size[1]:, -self.img_size[0]:, :] )

                if( self.mirror ):
                    # center
                    aug.append( np.fliplr( _img[h0:h1, w0:w1, :] ) )
                    # top left
                    aug.append( np.fliplr( _img[:self.img_size[1], :self.img_size[0], :] ) )
                    # top right
                    aug.append( np.fliplr( _img[:self.img_size[1], -self.img_size[0]:, :] ) )
                    # bottom left
                    aug.append( np.fliplr( _img[-self.img_size[1]:, :self.img_size[0], :] ) )
                    # bottom right
                    aug.append( np.fliplr( _img[-self.img_size[1]:, -self.img_size[0]:, :] ) )

        return aug

if( __name__ == '__main__' ):
    from scipy import misc
    face = misc.face()
    print( face.shape )

    cv2.imwrite( 'face.png', face )

    tencrop = TenCrop( [256] )
    aug = tencrop( [face] )
    for i, a in zip(range(len(aug)),aug):
        cv2.imwrite( '{}.png'.format(i), a )



import torch.nn as nn
import torchvision

class BuildNet:
    @classmethod
    def __prefix(cls):
        return 'build_'

    @classmethod
    def available_names(cls):
        prefix = cls.__prefix()
        names = []
        for name in dir(cls):
            if( name.startswith(prefix) ):
                names.append( name[len(prefix):] )
        return names

    @classmethod
    def build( cls, name ):
        if( not ( name in cls.available_names() ) ):
            raise NotImplementedError( '**' + name + '** is not implemented.')
        return eval(cls.__name__+'.'+cls.__prefix()+name)()


    '''
    build_[network name]():

    The output should be [N, ???, 1, 1] or [N, ???].
    ??? is the dimension of the feature

    The final layer may be
    nn.AdaptiveAvgPool2d(output_size=(1, 1)).

    '''

    @classmethod
    def build_alexnet( cls ):
        model = torchvision.models.alexnet(pretrained=True)
        modules=list(model.children())[:-2]
        modules.append( nn.AdaptiveAvgPool2d(output_size=(1, 1)) )
        network = nn.Sequential(*modules)
        return network, 256


    @classmethod
    def __vgg_p2p( cls, sequential ):
        k=1
        for module in sequential:
            if( isinstance( module, nn.modules.pooling.MaxPool2d) ):
                module.kernel_size=3
                module.stride=1
                module.padding=1
                k+=1
            elif( isinstance( module, nn.modules.conv.Conv2d ) ):
                module.dilation = (k,k)
                module.padding = (k,k)
        return sequential

    @classmethod
    def __vgg( cls, name, dim_feature, P2P ):
        model = eval('torchvision.models.'+name)(pretrained=True)
        modules = list(model.children())[:-2]
        if( P2P ):
            cls.__vgg_p2p(modules[0])
        else:
            modules.append( nn.AdaptiveAvgPool2d(output_size=(1, 1)) )
        network = nn.Sequential(*modules)
        return network, dim_feature

    @classmethod
    def build_vgg11( cls ):
        return cls.__vgg( 'vgg11', 512, False )

    @classmethod
    def build_vgg13( cls ):
        return cls.__vgg( 'vgg13', 512, False )

    @classmethod
    def build_vgg16( cls ):
        return cls.__vgg( 'vgg16', 512, False )

    @classmethod
    def build_vgg19( cls ):
        return cls.__vgg( 'vgg19', 512, False )

    @classmethod
    def build_vgg11p2p( cls ):
        return cls.__vgg( 'vgg11', 512, True )

    @classmethod
    def build_vgg13p2p( cls ):
        return cls.__vgg( 'vgg13', 512, True )

    @classmethod
    def build_vgg16p2p( cls ):
        return cls.__vgg( 'vgg16', 512, True )

    @classmethod
    def build_vgg19p2p( cls ):
        return cls.__vgg( 'vgg19', 512, True )

    @classmethod
    def __resnet( cls, name, dim_feature ):
        model = eval('torchvision.models.'+name)(pretrained=True)
        modules = list(model.children())[:-2]
        modules.append( nn.AdaptiveAvgPool2d(output_size=(1, 1)) )
        network = nn.Sequential(*modules)
        return network, dim_feature


    @classmethod
    def build_resnet18( cls ):
        return cls.__resnet( 'resnet18', 512 )

    @classmethod
    def build_resnet34( cls ):
        return cls.__resnet( 'resnet34', 512 )

    @classmethod
    def build_resnet101( cls ):
        return cls.__resnet( 'resnet101', 2048 )

    @classmethod
    def build_resnet152( cls ):
        return cls.__resnet( 'resnet152', 2048 )

    @classmethod
    def build_densenet121( cls, GAP=True ):
        model = torchvision.models.densenet121(pretrained=True)
        modules=list(model.children())[:-1]
        if( GAP ):
            modules.append( nn.AdaptiveAvgPool2d(output_size=(1, 1)) )
        network = nn.Sequential(*modules)
        return network, 1024

    @classmethod
    def build_densenet169( cls ):
        model = torchvision.models.densenet169(pretrained=True)
        modules=list(model.children())[:-1]
        modules.append( nn.AdaptiveAvgPool2d(output_size=(1, 1)) )
        network = nn.Sequential(*modules)
        return network, 1664

    @classmethod
    def build_densenet201( cls ):
        model = torchvision.models.densenet201(pretrained=True)
        modules=list(model.children())[:-1]
        modules.append( nn.AdaptiveAvgPool2d(output_size=(1, 1)) )
        network = nn.Sequential(*modules)
        return network, 1920

    @classmethod
    def build_densenet161( cls ):
        model = torchvision.models.densenet161(pretrained=True)
        modules=list(model.children())[:-1]
        modules.append( nn.AdaptiveAvgPool2d(output_size=(1, 1)) )
        network = nn.Sequential(*modules)
        return network, 2208

    @classmethod
    def build_googlenet( cls ):
        model = torchvision.models.googlenet(pretrained=True)
        modules=list(model.children())[:-3]
        modules.append( nn.AdaptiveAvgPool2d(output_size=(1, 1)) )
        network = nn.Sequential(*modules)
        return network, 1024

    @classmethod
    def build_mobilenet( cls ):
        model = torchvision.models.mobilenet_v2(pretrained=True)
        modules=list(model.children())[:-1]
        modules.append( nn.AdaptiveAvgPool2d(output_size=(1, 1)) )
        network = nn.Sequential(*modules)
        return network, 1280

if __name__=='__main__':
    print( BuildNet.available_names() )



from .buildnet import BuildNet
from .cnn import CNN
from .cnn import PixelFeature
from .aug import TenCrop
from .aug import Mirror
from . import antbee

__all__ = ['BuildNet', 'CNN','PixelFeature', 'TenCrop', 'Mirror', 'antbee']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

import numpy as np
import cv2
import torch

from img2feat import *

class tests( unittest.TestCase ):
###################################################################
    def suite():
        suite = unittest.TestSuite()
        suite.addTests(unittest.makeSuite(tests))
        return suite
###################################################################

    @classmethod
    def setUpClass(cls): # it is called before test starting
        pass

    @classmethod
    def tearDownClass(cls): # it is called before test ending
        pass

    def setUp(self): # it is called before each test
        pass

    def tearDown(self): # it is called after each test
        pass



###################################################################
    def test_buildnet(self):
        img = torch.rand( (1,3,224,224), dtype=torch.float32 )

#        names = BuildNet.available_names()
        names = ['vgg11', 'resnet18']
        for name in names:
            net, dim_feature = BuildNet.build( name )
            feat = net( img )
            self.assertEqual( feat.shape[0], 1 )
            self.assertEqual( feat.shape[1], dim_feature )
            self.assertEqual( feat.shape[2], 1 )
            self.assertEqual( feat.shape[3], 1 )

###################################################################
    def test_cnn(self):
        img = (np.random.rand( 224, 224, 3 )*255).astype(np.uint8)

#        names = BuildNet.available_names()
        names = ['vgg11', 'resnet18']
        for name in names:
            cnn = CNN( name )
            feat = cnn( [img,img] )
            self.assertEqual( feat.ndim, 2 )
            self.assertEqual( feat.shape[0], 2 )
            self.assertEqual( feat.shape[1], cnn.dim_feature )

###################################################################
    def test_pixelfeature(self):
        img = (np.random.rand( 128, 64, 3 )*255).astype(np.uint8)

#        names = BuildNet.available_names()
        names = ['vgg11']
        for name in names:
            pf = PixelFeature( name )
            f = pf( [img,img] )
            self.assertEqual( len(f), 2 )
            self.assertEqual( f[0].shape[0], img.shape[0] )
            self.assertEqual( f[0].shape[1], img.shape[1] )
            self.assertEqual( f[0].shape[2], pf.dim_feature )


###################################################################
    def test_aug(self):
        img = (np.random.rand( 224, 224, 3 )*255).astype(np.uint8)
        mirror = Mirror()
        aug = mirror( [img] )
        self.assertEqual( len(aug), mirror.nb_aug )
        for a in aug:
            self.assertEqual( a.shape[0], 224 )
            self.assertEqual( a.shape[1], 224 )

        img = (np.random.rand( 480, 640, 3 )*255).astype(np.uint8)
        tencrop = TenCrop( [256] )
        aug = tencrop( [img] )
        self.assertEqual( len(aug), tencrop.nb_aug )
        for a in aug:
            self.assertEqual( a.shape[0], 224 )
            self.assertEqual( a.shape[1], 224 )

###################################################################
    def test_antbee(self):
#        names = BuildNet.available_names()
        names = ['vgg11', 'resnet18']
        for name in names:
            cnn = CNN( name )
            (Xtrain, Ytrain), (Xtest, Ytest) = antbee.load_squared_npy( name )
            self.assertEqual( Ytrain.shape[0], 29+36 )
            self.assertEqual( Xtrain.shape[0], 29+36 )
            self.assertEqual( Xtrain.shape[1], cnn.dim_feature )
            self.assertEqual( Ytest.shape[0], 18+26 )
            self.assertEqual( Xtest.shape[0], 18+26 )
            self.assertEqual( Xtest.shape[1], cnn.dim_feature )

if( __name__ == '__main__' ):
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# convert README.md to README.rst
# % pandoc --from markdown --to rst README.md -o README.rst

# make distribution
# % python setup.py sdist bdist_wheel

# upload to pypi
# % twine upload --repository pypi dist/*

# uninstall
# % python setup.py install --record installed_files
# % cat installed_files | xargs rm -rf
# % rm installed_files

import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name= 'img2feat', # Application name:
    version= '0.2.0', # Version number

    author= 'Masayuki Tanaka', # Author name
    author_email= 'mastnk@gmail.com', # Author mail

    url='https://github.com/mastnk/img2feat', # Details
    description='Convert image to feature based on convolutional neural network (CNN).', # short description
    long_description=read('README.rst'), # long description
    install_requires=[ # Dependent packages (distributions)
      'scikit-learn', 'numpy', 'torch', 'torchvision','opencv-python', 'scikit-build', 'cmake', 'scipy',
    ],

    include_package_data=False, # Include additional files into the package
    packages=find_packages(exclude=('tests')),

    test_suite = 'tests',

    classifiers=[
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Programming Language :: Python :: 3.6',
    ]
)




import cv2
import numpy as np
import torch

from img2feat import BuildNet

class CNN:
    @classmethod
    def available_networks(cls):
        names = []
        for name in BuildNet.available_names():
            if( not name.endswith( 'p2p' ) ):
                names.append( name )
        return names

    def __init__( self, name='vgg11', gpu=False, img_size=(224,224), batch_size=128 ): # img_size=(width,hegith)
        self.__gpu = gpu
        self.__img_size=img_size
        self.batch_size = batch_size

        if( name.lower() in self.available_networks() ):
            self.__network, self.__dim_feature = BuildNet.build( name.lower() )
        else:
            raise NotImplementedError( '**' + name + '** is not implemented.')

        for p in self.__network.parameters():
            p.requires_grad = False

        if self.__gpu:
            self.__network.cuda()
        self.__network.eval()

    @property
    def img_size( self ):
        return self.__img_size

    @property
    def dim_feature( self ):
        return self.__dim_feature

    def imgs2tensor(self, imgs):
        # This function transfer the image to tensor.
        # imgs is list of img
        # img shape : (width, height, channel)
        # tensor shape : (batch, channel, height, width)
        # cv2 (img):BGR, pytorch(tensor):RGB
        # self.img_size: (width, height)

        nb_imgs = len(imgs)
        img_np = np.zeros( (nb_imgs, 3, self.__img_size[1], self.__img_size[0]), dtype=np.float32 )

        for i, img in zip(range(nb_imgs), imgs):
            _img = cv2.resize( img, self.__img_size )

            # BGR -> RGB and fit to imagenet pre-trained model
            img_np[i,2,:,:] = (_img[:,:,0]/255-0.485)/0.229
            img_np[i,1,:,:] = (_img[:,:,1]/255-0.456)/0.224
            img_np[i,0,:,:] = (_img[:,:,2]/255-0.406)/0.225

        img_tensor = torch.from_numpy( img_np )
        if( self.__gpu ):
            img_tensor = img_tensor.cuda()

        return img_tensor

    def tensor2array(self, tensor):
        if( self.__gpu ):
            array = tensor.data.cpu().numpy()
        else:
            array = tensor.data.numpy()
        return array.squeeze( (2,3) )

    def imgs2feature( self, imgs ):
        tensor_imgs = self.imgs2tensor(imgs)
        tensor_feature = self.__network(tensor_imgs)
        array_feature = self.tensor2array( tensor_feature )
        return array_feature

    def __call__( self, imgs ):
        if( type(imgs) is not list ):
            imgs = [imgs]

        array_feature = np.zeros( (len(imgs), self.dim_feature), dtype=np.float32 )

        N = len(imgs) // self.batch_size
        for i in range(N):
            array_feature[i*self.batch_size:(i+1)*self.batch_size,:] = \
                self.imgs2feature( imgs[i*self.batch_size:(i+1)*self.batch_size] )

        array_feature[N*self.batch_size:,:] = \
            self.imgs2feature( imgs[N*self.batch_size:] )

        return array_feature

class PixelFeature:
    @classmethod
    def available_networks(cls):
        names = []
        for name in BuildNet.available_names():
            if( name.endswith( 'p2p' ) ):
                names.append( name[:-3] )
        return names

    def __init__( self, name='vgg11', gpu=False ):
        self.__gpu = gpu

        if( name.lower() in self.available_networks() ):
            self.__network, self.__dim_feature = BuildNet.build( name.lower()+'p2p' )
        else:
            raise NotImplementedError( '**' + name + '** is not implemented.')

        for p in self.__network.parameters():
            p.requires_grad = False

        if self.__gpu:
            self.__network.cuda()
        self.__network.eval()

    @property
    def dim_feature( self ):
        return self.__dim_feature

    def img2tensor(self, img):
        # This function transfer the image to tensor.
        # imgs is list of img
        # img shape : (width, height, channel)
        # tensor shape : (batch, channel, height, width)
        # cv2 (img):BGR, pytorch(tensor):RGB
        # self.img_size: (width, height)

        img_np = np.zeros( (1, 3, img.shape[0], img.shape[1]), dtype=np.float32 )

        # BGR -> RGB and fit to imagenet pre-trained model
        img_np[0,2,:,:] = (img[:,:,0]/255-0.485)/0.229
        img_np[0,1,:,:] = (img[:,:,1]/255-0.456)/0.224
        img_np[0,0,:,:] = (img[:,:,2]/255-0.406)/0.225

        img_tensor = torch.from_numpy( img_np )
        if( self.__gpu ):
            img_tensor = img_tensor.cuda()

        return img_tensor

    def tensor2array(self, tensor):
        if( self.__gpu ):
            array = tensor.data.cpu().numpy()
        else:
            array = tensor.data.numpy()
        return array.transpose( (0,2,3,1) ).squeeze( 0 )

    def __call__( self, imgs ):
        if( type(imgs) is not list ):
            imgs = [imgs]

        array_features = []
        for img in imgs:
            tensor_imgs = self.img2tensor(img)
            tensor_feature = self.__network(tensor_imgs)
            array_features.append( self.tensor2array( tensor_feature ) )
        return array_features




# https://aidiary.hatenablog.com/entry/20180217/1518833659
# https://pytorch.org/tutorials/beginner/transfer_learning_tutorial.html

import cv2
import numpy as np
import urllib.request
import os
import zipfile
import glob
import hashlib
import shutil

from img2feat import CNN

str = ['ant', 'bee']

def load_hymenoptera( root ):
    filename = 'hymenoptera_data.zip'
    url = 'https://download.pytorch.org/tutorial/'+filename

    dir = os.path.join( root, 'ant-vs-bee' )
    os.makedirs( dir, exist_ok=True )
    dir_filename = os.path.join( dir, filename )

    download = False
    if( os.path.exists( dir_filename ) ):
        with open( dir_filename, 'rb' ) as fin:
            md5 = hashlib.md5( fin.read() ).hexdigest()
        if( md5 != '5f8c32a6554f6acb4d649776e7735e48' ):
            download = True
    else:
        download = True

    if( download ):
        urllib.request.urlretrieve(url, dir_filename)

    extract = False
    nb_files = [124, 121, 70, 83]
    dirs = ['train/ants/*', 'train/bees/*', 'val/ants/*', 'val/bees/*']
    for n, d in zip( nb_files, dirs ):
        files = glob.glob( os.path.join( dir, 'hymenoptera_data/'+d ) )
        if( len(files) != n ):
            extract = True

    if( extract ):
        if( os.path.exists( os.path.join( dir, 'hymenoptera_data' ) ) ):
            shutil.rmtree( os.path.join( dir, 'hymenoptera_data' ) )
        with zipfile.ZipFile(dir_filename) as existing_zip:
            existing_zip.extractall( dir )

def load_imgs( dir ):
    keys = [ 'train0', 'train1', 'test0', 'test1' ]
    dirs = ['train/ants/*', 'train/bees/*', 'val/ants/*', 'val/bees/*']
    data = {}

    for k,d in zip(keys, dirs):
        data[k] = []
        files = glob.glob( os.path.join( dir, d ) )
        files.sort()
        for file in files:
            img = cv2.imread( file )
            if( img is not None ):
                data[k].append( img )

    Xtrain = data['train0'] + data['train1']
    Ytrain = np.concatenate( (np.zeros( (len(data['train0']),), dtype=np.int ),
                              np.ones ( (len(data['train1']),), dtype=np.int ) ),
                              axis=0 )
    Xtest = data['test0'] + data['test1']
    Ytest = np.concatenate( (np.zeros( (len(data['test0']),), dtype=np.int ),
                             np.ones ( (len(data['test1']),), dtype=np.int ) ),
                              axis=0 )

    return (Xtrain, Ytrain), (Xtest, Ytest)


def extract_squared_imgs( dir ):
    org_dir = os.path.join( dir, 'hymenoptera_data' )
    sqr_dir = os.path.join( dir, 'hymenoptera_squared' )

    keys = [ 'train0', 'train1', 'test0', 'test1' ]
    dirs = ['train/ants/', 'train/bees/', 'val/ants/', 'val/bees/']

    extract = False
    nb_files = [29, 36, 18, 26]
    for n, d in zip( nb_files, dirs ):
        files = glob.glob( os.path.join( dir, 'hymenoptera_squared/'+d+'*' ) )
        if( len(files) != n ):
            extract = True

    if( extract ):
        if( os.path.exists(sqr_dir ) ):
            shutil.rmtree( sqr_dir )

        for k,d in zip(keys, dirs):
            dst_dir = os.path.join( sqr_dir, d )
            os.makedirs( dst_dir )
            files = glob.glob( os.path.join( org_dir, d )+'*' )
            files.sort()
            for file in files:
                img = cv2.imread( file )
                if( img is not None ):
                    height, width, _ = img.shape
                    ratio = height / width
                    if( ratio > 0.75 and ratio < 1.25 ):
                        if height >= width:
                            cropped_img = img[ int(height/2-width/2):int(height/2+width/2), :, : ]
                        else:
                            cropped_img = img[ :, int(width/2-height/2):int(width/2+height/2), : ]

                        img = cv2.resize( cropped_img, (256,256) )
                        dst_file = os.path.join( dst_dir, os.path.basename(file) )
                        cv2.imwrite( dst_file, img )

def load( squared=True, root=None ):
    if( root is None ):
        root = os.path.dirname(__file__)
    load_hymenoptera( root )

    dir = os.path.join( root, 'ant-vs-bee' )
    if( squared ):
        extract_squared_imgs( dir )
        dir = os.path.join( dir, 'hymenoptera_squared' )
    else:
        dir = os.path.join( dir, 'hymenoptera_data' )

    return load_imgs( dir )

def load_squared_npy( name, root=None ):
    if( root is None ):
        root = os.path.dirname(__file__)

    dir = os.path.join( root, 'ant-vs-bee/hymenoptera_squared' )
    dir_filename = os.path.join( dir, name+'.npz' )

    if( not os.path.exists( dir_filename ) ):
        (Xtrain, Ytrain), (Xtest, Ytest) = load( True, root )
        cnn = CNN(name)
        Xtrain = cnn( Xtrain )
        Xtest = cnn( Xtest )
        np.savez( dir_filename, Xtrain=Xtrain, Ytrain=Ytrain, Xtest=Xtest, Ytest=Ytest )

    d = np.load( dir_filename )
    return (d['Xtrain'], d['Ytrain']), (d['Xtest'], d['Ytest'])

